{***************************************************************
 *
 * Unit Name: AJB_ADO
 * Purpose  : Extend the functionality of ADO components
 * Author   : Andrew Baylis
 * Date     : 16/06/2001
 * History  : Ver 1.0 - Extend ADOConnection to include DatabaseName
              field so that it can be easily changed at runtime.
              Ver 2.0 - Added Provider settings
              Ver 2.1 - Get / Set QueryCommandText
                        GetADOColumn Info without opening table
                        DropTable
                        ChangeLinks for linked tables
                        Encrypt / decrypt string & memo fields
              Ver 2.2 - Add access to ADO errors
              Ver 2.3 - ListQueries,ListTables,ListProcedures
              Ver 2.4 - Get field type for a fieldname from a query/table
              Ver 2.5 - Fill the fielddefs of another dataset with info
                        from a given query / table. does this without opening
                        the table.
              Ver 2.6 -  Handling of Disconnected Recordsets
                         (CursorType=ctDynamic/ctKeyset / locktype=ltBatchOptimistic
                         DisconnectRS(tab: TADODataset) - disconnects an open dataset
                         OpenAndDisconnectRS(tab: TADODataset)

                         ReconnectRS(tab: TADODataset; trapErrs: Boolean = True; showDlg: Boolean = false)
                         trapErrs will catch errors due to other users changing data

                         ReloadRS(tab: TADODataset); - reloads dataset. Is also used in Reconnect

                         SaveAndDisconnectRS(tab: TADODataset;showDlg:Boolean = True);
              Ver 2.7 - Added ISAM property - allows any file with an installable ISAM
                        to be read with ADO Jet engine. eg: Excel, DBase, FoxPro
              Ver 2.8 - TableExists, FieldExists functions - quicker than opening the table
                        or query directly
              Ver 2.9 - Add/remove field from table
              Ver 3.0 - Better handling of disconnected recordsets (old way would keep an active connection)
                        FCatalog is only created when needed and then discarded, thus preventing
                        a connection remaining open. Now when Connected=false, it is guaranteed that
                        there are no active connections to the database
 *
 ****************************************************************}

unit AJB_ADO;

interface

uses Windows, classes, ADOdb, ADO_Func, sysutils, AJB_ADOX,
  ADOInt, ADOX_TLB, DB, oledb, inifiles, comobj, registry,
  FileCtrl, Variants;

type
  TProvstring = string;
  TDriverstring = string;
  TISAMString = string;
  TFilterType = (fltAccess, fltExcel, fltLotus, fltText);

  RColInfo = record
    FieldName: string;
    FieldType: TFieldType;
    FieldSize: Integer;
    FieldFlags: Integer;
    FieldPrecision: Integer;
    IsNullable: Boolean;
  end;
  PColInfo = ^RColInfo;

  TADOColumnList = class(TObject)
  private
    FList: TList;
    function GetItems(Index: Integer): PColInfo;
    function GetCount: Integer;
  public
    destructor Destroy; override;
    procedure Clear;
    constructor Create;
    function Add: PColInfo;
    function GetFieldInfo(FieldName: string): PColInfo;
    property Items[Index: Integer]: PColInfo read GetItems; default;
    property Count: Integer read GetCount;
  end;

  TSchemaItem = (itTable, itView, itProcedure, itSystem, itUsers, itLink);
  TSchemaItemSet = set of TSchemaItem;

  TAJB_ADOConnection = class(TADOConnection)
  private
    FCheckReadOnly: Boolean;
    FConnectionList: TStringList;
    function GetADOErrors: Errors;
    function GetDatabaseName: TFileName;
    function GetEncrypt: Boolean;
    function GetErrorCount: Integer;
    function GetISAM: TISAMstring;
    function GetModeString: string;
    function GetPassword: string;
    function GetProvider: TProvstring;
    function GetVersion: string;
    procedure SetDatabaseName(const Value: TFilename);
    procedure SetEncrypt(Value: Boolean);
    procedure SetISAM(Value: TISAMstring);
    procedure SetModeString(const Value: string);
    procedure SetPassword(const Value: string);
    procedure SetProvider(const Value: TProvstring);
    procedure SetVersion(const Value: string);
  protected
    procedure DoConnect; override;
    procedure GetItemList(which: TSchemaItemSet; List: TStrings);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function AddField(const TableName: string; fldDef: TFieldDef; AllowZeroLength:
      Boolean = True): Boolean;
    procedure ChangeAllLinks(const NewDatabase: string);
    function ChangeCommandText(const QueryName: string; const NewText: string):
      Boolean;
    function ChangeFieldDef(const TableName, Fieldname: string; newDef: TFieldDef;
      AllowZeroLength: Boolean = True): Boolean;
    procedure ClearErrors;
    function DeleteField(const TableName, FieldName: string): Boolean;
    function DropTable(const TableName: string): Boolean;
    function FieldExists(const TableName, FieldName: string): Boolean;
    procedure GetADOColumnList(const Tablename: string; List: TADOColumnList;
      NamesList: TStringList = nil);
    procedure GetADOErrorList(List: TStrings);
    procedure GetADOFieldNames(const Tablename: string; List: TStringList);
    function GetFieldType(tabname, fieldname: string): TFieldType;
    procedure GetProcedureList(sl: TStrings);
    function GetQueryCommandText(const QueryName: string): string;
    procedure GetTableList(sl: TStrings);
    procedure GetTableNames(List: TStrings; SystemTables: Boolean = False);
    procedure GetUserList(sl: TStrings);
    procedure GetViewList(sl: TStrings);
    procedure MakeDatasetAs(Dest: TDataset; query: string);
    function TableExists(const TableName: string): Boolean;
    property ADOErrors: Errors read GetADOErrors;
    property ErrorCount: Integer read GetErrorCount;
    property ModeString: string read GetModeString write SetModeString;
  published
    property CheckReadOnly: Boolean read FCheckReadOnly write FCheckReadOnly;
    property Database: TFilename read GetDatabaseName write SetDatabaseName;
    property Encrypt: Boolean read GetEncrypt write SetEncrypt;
    property ISAM: TISAMstring read GetISAM write SetISAM;
    property Password: string read GetPassword write SetPassword;
    property Provider: TProvstring read GetProvider write SetProvider;
    property Version: string read GetVersion write SetVersion;
  end;

  TODBC_Connection = class(TADOConnection)
  private
    FODBCServer: string;
    FDatabaseFile: TFilename;
    FODBCDriver: TDriverstring;
    function GetProvider: WideString;
    function GetConnectionString: string;
  protected
    procedure SetConnected(Value: Boolean); override;
    function GetConnected: Boolean; override;
    procedure MakeConnectionString;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property ODBCDriver: TDriverstring read FODBCDriver write FODBCDriver;
    property ODBCServer: string read FODBCServer write FODBCServer;
    property Provider read GetProvider;
    property ConnectionString: string read GetConnectionString;
    property DatabaseFile: TFilename read FDatabaseFile write FDatabaseFile;
    property Connected read GetConnected write SetConnected;
  end;

procedure DisconnectRS(tab: TADODataset);

procedure OpenAndDisconnectRS(tab: TADODataset);

procedure ReloadRS(tab: TADODataset);

procedure ReconnectRS(tab: TADODataset; conn: TADOConnection; trapErrs: Boolean = True; showDlg:
  Boolean = False);

procedure SaveAndDisconnectRS(tab: TADODataset; conn: TADOConnection;
  trapErrs: Boolean = True; showDlg: Boolean = True);

procedure RefreshAndDisconnectRS(tab: TADODataset; conn: TADOConnection);

implementation

uses TypInfo, Consts, SQLEdit, Dialogs, AdoConEd;

const
  cPassword = 'Password';
  versionStr = 'AJB Ver 3.0';
  eRowError: HRESULT = -2147217864;
  sError1 = 'Cannot update %d records as they have ' + #13#10 + 'been changed by another user';
  noISAM = '(Access)';

procedure OpenAndDisconnectRS(tab: TADODataset);
begin
  Tab.Open;
  DisconnectRS(Tab);
end;

procedure ReloadRS(tab: TADODataset);
begin
  if not Tab.Active then Exit;
  Tab.DisableControls;
  try
    //can't requery as connection has changed
    Tab.Active := False;
    Tab.Active := True;
  finally
    Tab.EnableControls;
  end;
end;

procedure DisconnectRS(tab: TADODataset);
begin
  Tab.Connection := nil;
  Tab.First; //reset cursor
end;

procedure ReconnectRS(tab: TADODataset; conn: TADOConnection; trapErrs: Boolean = True; showDlg:
  Boolean = False);
var
  i: Integer;
begin
  if not (conn.Connected) then conn.connected := True; //ensure connection is open
  Tab.Connection := conn;
  if trapErrs then
    begin
      try
        try
          if (Tab.Active) and (Tab.LockType = ltBatchOptimistic) then Tab.UpdateBatch;
        except
          on e: EOleException do
            begin
              if (e.ErrorCode = eRowError) then
                begin
                  i := 0;
                  if showDlg then
                    begin
                      Tab.DisableControls;
                      Tab.First;
                      while not Tab.Eof do
                        begin
                          if (rsModified in Tab.RecordStatus) then
                            Inc(i);
                          Tab.Next;
                        end;
                      Tab.EnableControls;
                      MessageDlg(Format(sError1, [i]), mtError, [mbOK], 0);
                    end;
                end
              else
                MessageDlg(IntToStr(e.errorcode) + #13#10 + e.Message, mtWarning, [mbOK], 0);
            end;
        end; //except
      finally
        if (Tab.Active) and (Tab.LockType = ltBatchOptimistic) then ReloadRS(Tab);
      end;
    end
  else if (Tab.Active) and (Tab.LockType = ltBatchOptimistic) then
    Tab.UpdateBatch;
end;

procedure SaveAndDisconnectRS(tab: TADODataset; conn: TADOConnection;
  trapErrs: Boolean = True; showDlg: Boolean = True);
begin
  if (Tab.Active) and (Tab.LockType = ltBatchOptimistic) then
    begin
      Tab.DisableControls;
      try
        ReconnectRS(Tab, conn, trapErrs, showDlg);
        DisconnectRS(Tab);
      finally
        Tab.EnableControls;
      end;
    end;
end;

procedure RefreshAndDisconnectRS(tab: TADODataset; conn: TADOConnection);
begin
  if not (conn.Connected) then conn.connected := True; //ensure connection is open
  Tab.Connection := conn;
  if (Tab.Active) and (Tab.LockType = ltBatchOptimistic) then
    Tab.CancelBatch();
  ReloadRS(Tab);
  DisconnectRS(Tab);
end;

constructor TAJB_ADOConnection.Create(AOwner: TComponent);
begin
  inherited;
  FConnectionList := TStringList.Create;
  Provider := Jet4;
  LoginPrompt := False;
  FCheckReadOnly := False;
  ISAM := noISAM;
end;

destructor TAJB_ADOConnection.Destroy;
begin
  FConnectionList.Free;
  inherited;
end;

function TAJB_ADOConnection.AddField(const TableName: string; fldDef: TFieldDef;
  AllowZeroLength: Boolean = True): Boolean;
var
  mTab: _Table;
  i: Integer;
  s: string;
  mCol: IADOXColumn;
  FCatalog: _Catalog;
begin
  FCatalog := CoCatalog.Create;
  FCatalog._Set_ActiveConnection(ConnectionObject);
  try
    Result := False;
    FCatalog.Tables.Refresh;
    i := 0;
    while (not Result) and (i < FCatalog.Tables.Count) do
      begin
        s := FCatalog.Tables.Item[i].Name;
        if (StrIComp(PChar(s), PChar(TableName)) = 0) then
          begin
            Result := True;
            mTab := FCatalog.Tables.Item[i];
          end;
        inc(i);
      end;

    if Result then
      begin
        mCol := TADOXColumn.Create(FCatalog); // sets ParentCatalog as well
        mCol.Name := fldDef.Name;
        mCol.DataType := fldDef.DataType;
        mCol.Size := fldDef.Size;
        mCol.Precision := fldDef.Precision;
        if (fldDef.DataType in [ftString, ftWideString]) then
          mCol.AllowZeroLength := AllowZeroLength;
        mCol.Required := fldDef.Required;
        mTab.Columns.Append(mCol.ADOColumn, Unassigned, Unassigned);
        FCatalog.Tables.Refresh;
      end;
  finally
    FCatalog := nil;
  end;
end;

procedure TAJB_ADOConnection.ChangeAllLinks(const NewDatabase: string);
var
  i: Integer;
  needrefresh: Boolean;
  FCatalog: _Catalog;
begin
  needrefresh := False;
  if not Connected then Exit;
  try
    FCatalog := CoCatalog.Create;
    FCatalog._Set_ActiveConnection(ConnectionObject);
    for i := 0 to FCatalog.Tables.Count - 1 do
      if (FCatalog.Tables[i].Type_ = 'LINK') and
        (FCatalog.Tables[i].Properties[cJETLinkSource].Value <> NewDatabase) then
        begin
          FCatalog.Tables[i].Properties[cJETLinkSource].Value := NewDatabase;
          needrefresh := True;
        end;

  finally
    if needrefresh then
      FCatalog.Tables.Refresh;
    FCatalog := nil;
  end;
end;

function TAJB_ADOConnection.ChangeCommandText(const QueryName: string; const
  NewText: string): Boolean;
var
  i: Integer;
  v: View;
  p: Procedure_;
  //cmd:OleVariant;
  cmd: _Command;
  s: string;
  FCatalog: _Catalog;
begin
  Result := false;
  CheckActive;
  FCatalog := CoCatalog.Create;
  FCatalog._Set_ActiveConnection(ConnectionObject);
  try
    cmd := CoCommand.Create;
    Cmd.CommandText := NewText;
    Cmd.CommandType := adCmdText;
    Cmd.Name := QueryName;

    for i := 0 to FCatalog.Views.Count - 1 do
      begin
        s := FCatalog.Views.Item[i].Name;
        if (StrIComp(PChar(s), PChar(QueryName)) = 0) then
          begin
            Result := True;
            v := FCatalog.Views.Item[i];
            Break;
          end;
      end;
    if Result then
      begin
        v._Set_Command(cmd);
        FCatalog.Views.Refresh;
      end
    else
      begin
        for i := 0 to FCatalog.Procedures.Count - 1 do
          begin
            s := FCatalog.Procedures.Item[i].Name;
            if (StrIComp(PChar(s), PChar(QueryName)) = 0) then
              begin
                Result := True;
                p := FCatalog.Procedures.Item[i];
                Break;
              end;
          end;
        if Result then
          begin
            p._Set_Command(cmd);
            FCatalog.Procedures.Refresh;
          end;
      end;
    FCatalog.Tables.Refresh;
  finally
    FCatalog := nil;
  end;
end;

function TAJB_ADOConnection.ChangeFieldDef(const TableName, Fieldname: string;
  newDef: TFieldDef; AllowZeroLength: Boolean = True): Boolean;
var
  mTab: _Table;
  i: Integer;
  s: string;
  mCol: IADOXColumn;
  FCatalog: _Catalog;
begin
  FCatalog := CoCatalog.Create;
  FCatalog._Set_ActiveConnection(ConnectionObject);
  try
    Result := False;
    FCatalog.Tables.Refresh;
    i := 0;
    while (not Result) and (i < FCatalog.Tables.Count) do
      begin
        s := FCatalog.Tables.Item[i].Name;
        if (StrIComp(PChar(s), PChar(TableName)) = 0) then
          begin
            Result := True;
            mTab := FCatalog.Tables.Item[i];
          end;
        inc(i);
      end;

    if Result then
      begin
        mCol := TADOXColumn.Create(FCatalog); // sets ParentCatalog as well
        Result := False;
        while (not Result) and (i < mTab.Columns.Count) do
          begin
            s := mTab.Columns.Item[i].Name;
            if (StrIComp(PChar(s), PChar(FieldName)) = 0) then
              begin
                Result := True;
                mCol.ADOColumn := mTab.Columns[i];
              end;
            inc(i);
          end;
        if Result then
          begin
            mCol.Name := newDef.Name;
            mCol.DataType := newDef.DataType;
            mCol.Size := newDef.Size;
            mCol.Required := newDef.Required;
            mCol.Precision := newDef.Precision;
            if (newDef.DataType in [ftString, ftWideString]) then
              mCol.AllowZeroLength := AllowZeroLength;
            mTab.Columns.Refresh;
          end;
        FCatalog.Tables.Refresh;
      end;
  finally
    FCatalog := nil;
  end;
end;

procedure TAJB_ADOConnection.ClearErrors;
begin
  if Connected then
    ConnectionObject.Errors.Clear;
end;

function TAJB_ADOConnection.DeleteField(const TableName, FieldName: string):
  Boolean;
var
  mTab: _Table;
  i: Integer;
  s: string;
  FCatalog: _Catalog;
begin
  FCatalog := CoCatalog.Create;
  FCatalog._Set_ActiveConnection(ConnectionObject);
  try
    Result := False;
    FCatalog.Tables.Refresh;
    i := 0;
    while (not Result) and (i < FCatalog.Tables.Count) do
      begin
        s := FCatalog.Tables.Item[i].Name;
        if (StrIComp(PChar(s), PChar(TableName)) = 0) then
          begin
            Result := True;
            mTab := FCatalog.Tables.Item[i];
          end;
        inc(i);
      end;

    if Result then
      begin
        Result := False;
        i := 0;
        while (not Result) and (i < mTab.Columns.Count) do
          begin
            s := mTab.Columns.Item[i].Name;
            if (StrIComp(PChar(s), PChar(FieldName)) = 0) then
              begin
                Result := True;
                mTab.Columns.Delete(i);
              end;
            inc(i);
          end;
        FCatalog.Tables.Refresh;
      end;
  finally
    FCatalog := nil;
  end;
end;

procedure TAJB_ADOConnection.DoConnect;
var
  FileName: string;
  attrs: Integer;
begin
  if FCheckReadOnly then
    begin
      FileName := Database;
      if (FileName <> '') then
        begin
          attrs := FileGetAttr(FileName);
          if (attrs <> -1) and ((attrs and $00000001) > 0) then
            begin
              Mode := cmShareDenyWrite;
              SetModeString('Share Deny Write');
            end;
        end;
    end;

  inherited;
end;

function TAJB_ADOConnection.DropTable(const TableName: string): Boolean;
var
  i: Integer;
  s: string;
  FCatalog: _Catalog;
begin
  Result := false;
  CheckActive;
  FCatalog := CoCatalog.Create;
  FCatalog._Set_ActiveConnection(ConnectionObject);
  try
    FCatalog.Tables.Refresh;
    if not Result then
      for i := 0 to FCatalog.Tables.Count - 1 do
        begin
          s := FCatalog.Tables.Item[i].Name;
          if (StrIComp(PChar(s), PChar(TableName)) = 0) then
            begin
              Result := True;
              FCatalog.Tables.Delete(i);
              Break;
            end;
        end;

    FCatalog.Views.Refresh;
    if not Result then
      for i := 0 to FCatalog.Views.Count - 1 do
        begin
          s := FCatalog.Views.Item[i].Name;
          if (StrIComp(PChar(s), PChar(TableName)) = 0) then
            begin
              Result := True;
              FCatalog.Views.Delete(i);
              Break;
            end;
        end;

    FCatalog.Procedures.Refresh;
    if not Result then
      for i := 0 to FCatalog.Procedures.Count - 1 do
        begin
          s := FCatalog.Procedures.Item[i].Name;
          if (StrIComp(PChar(s), PChar(TableName)) = 0) then
            begin
              Result := True;
              FCatalog.Procedures.Delete(i);
              Break;
            end;
        end;

    FCatalog.Tables.Refresh;
    FCatalog.Views.Refresh;
    FCatalog.Procedures.Refresh;
  finally
    FCatalog := nil;
  end;
end;

function TAJB_ADOConnection.FieldExists(const TableName, FieldName: string):
  Boolean;
var
  sl: TStringList;
  i: Integer;
begin
  sl := TStringList.Create;
  try
    GetADOFieldNames(TableName, sl);
    sl.Sorted := True;
    Result := sl.find(FieldName, i);
  finally
    sl.Free;
  end;
end;

procedure TAJB_ADOConnection.GetADOColumnList(const Tablename: string; List:
  TADOColumnList; NamesList: TStringList = nil);

const
  DATA_TYPE = 'DATA_TYPE';
  COLUMN_FLAGS = 'COLUMN_FLAGS';
  IS_NULLABLE = 'IS_NULLABLE';
  FIELD_SIZE = 'CHARACTER_MAXIMUM_LENGTH';
  NUM_PRECISION = 'NUMERIC_PRECISION';
  COLUMN_NAME = 'COLUMN_NAME'; { Do not localize }

var
  Fields: _Recordset;
  recinfo: PColInfo;
  dt, ff: Integer;
  ft: TFieldType;

begin
  CheckActive;
  if (List <> nil) then
    List.Clear
  else if (NamesList <> nil) then
    NamesList.Clear;
  if Length(Trim(Tablename)) = 0 then
    Exit;

  try
    Fields := ConnectionObject.OpenSchema(adSchemaColumns, VarArrayOf([Null, Null, TableName]),
      EmptyParam);
    if Fields = nil then Exit;
    Fields.MoveFirst;
    while not Fields.EOF do
      begin
        if (List <> nil) then
          begin
            recinfo := List.Add;
            if recinfo <> nil then
              begin
                recinfo.FieldName := VarToStr(Fields.Fields[COLUMN_NAME].Value);
                recinfo.FieldSize := StrToIntDef(VarToStr(Fields.Fields[FIELD_SIZE].Value), 0);
                recinfo.FieldFlags := StrToIntDef(VarToStr(Fields.Fields[COLUMN_FLAGS].Value), 0);
                recinfo.FieldPrecision := StrToIntDef(VarToStr(Fields.Fields[NUM_PRECISION].Value), 0);
                recinfo.IsNullable := ((Fields.Fields[IS_NULLABLE].Value) = True);
                dt := StrToIntDef(VarToStr(Fields.Fields[DATA_TYPE].Value), 0);
                ff := recinfo.FieldFlags;
                case dt of
                  DBTYPE_WSTR:
                    if ((ff and DBCOLUMNFLAGS_ISLONG) > 0) then
                      ft := ftMemo
                    else
                      ft := ftString;
                  DBTYPE_BYTES: ft := ftBlob;
                  DBTYPE_I4:
                    if ((ff and DBCOLUMNFLAGS_ISNULLABLE) > 0) then
                      ft := ftInteger
                    else
                      ft := ftAutoInc;
                  DBTYPE_I2: ft := ftSmallint;
                  DBTYPE_R4, DBTYPE_R8: ft := ftFloat;
                  DBTYPE_DATE: ft := ftDateTime;
                  DBTYPE_CY: ft := ftCurrency;
                  DBTYPE_UI1: ft := ftWord;
                  else
                    ft := ftUnknown;
                end; // case
                recinfo.FieldType := ft;
              end;
          end
        else if (NamesList <> nil) then
          NamesList.Add(VarToStr(Fields.Fields[COLUMN_NAME].Value));

        Fields.MoveNext;
      end;
  except
    // don't do anything. If tablename doesn't exist, List is empty
    Fields := nil;
  end;
end;

procedure TAJB_ADOConnection.GetADOErrorList(List: TStrings);
var
  i, c: Integer;
begin
  List.Clear;
  c := ErrorCount;
  if (c > 0) then
    for i := 0 to c - 1 do
      List.Add(ConnectionObject.Errors[i].Description);
end;

function TAJB_ADOConnection.GetADOErrors: Errors;
begin
  if Connected then
    Result := ConnectionObject.Errors
  else
    Result := nil;
end;

procedure TAJB_ADOConnection.GetADOFieldNames(const Tablename: string; List:
  TStringList);
begin
  GetADOColumnList(TableName, nil, List);
end;

function TAJB_ADOConnection.GetDatabaseName: TFileName;
begin
  if ConnectionString <> '' then
    begin
      ConvertStringToList(ConnectionString, FConnectionList);
      Result := FConnectionList.Values['Data Source'];
    end
  else
    Result := '';
end;

function TAJB_ADOConnection.GetEncrypt: Boolean;
begin
  if ConnectionString <> '' then
    begin
      ConvertStringToList(ConnectionString, FConnectionList);
      Result := (FConnectionList.Values[cJETEncrypt] = 'True');
    end
  else
    Result := False;
end;

function TAJB_ADOConnection.GetErrorCount: Integer;
begin
  if Connected then
    Result := ConnectionObject.Errors.Count
  else
    Result := -1;
end;

function TAJB_ADOConnection.GetFieldType(tabname, fieldname: string): TFieldType;
var
  collist: TADOColumnList;
  p: PColInfo;
begin
  collist := TADOColumnList.Create;
  Result := ftUnknown;
  try
    GetADOColumnList(tabname, collist);
    p := collist.GetFieldInfo(FieldName);
    if (p <> nil) then
      Result := p.FieldType;
  finally
    collist.Free;
  end;
end;

function TAJB_ADOConnection.GetISAM: TISAMstring;
begin
  if ConnectionString <> '' then
    begin
      ConvertStringToList(ConnectionString, FConnectionList);
      Result := FConnectionList.Values['Extended Properties'];
      if (Result = '') or (Result = '""') then Result := noISAM;
    end
  else
    Result := noISAM;
end;

procedure TAJB_ADOConnection.GetItemList(which: TSchemaItemSet; List: TStrings);
var
  TypeField,
    NameField: TField;
  TableType: string;
  SchemaParam: OleVariant;
  DataSet: TADODataSet;
begin
  if not (Connected and Assigned(List)) then
    Exit; //safety check

  DataSet := TADODataSet.Create(nil);
  List.BeginUpdate; //NB nested calls are OK
  try
    List.Clear;
    if (itProcedure in which) then
      getProcedurenames(List)
    else if (itUsers in which) then
      begin
        //get list of users
        //Fields are COMPUTER_NAME,LOGIN_NAME,CONNECTED,SUSPECT_STATE
        SchemaParam := '{947bb102-5d43-11d1-bdbf-00c04fb92675}';
        OpenSchema(siProviderSpecific, EmptyParam, SchemaParam, DataSet);
        DataSet.First;
        while not DataSet.EOF do
          begin
            List.Add(DataSet.Fields[0].AsString + ',' +
              DataSet.Fields[1].AsString + ',' +
              DataSet.Fields[2].AsString + ',' +
              DataSet.Fields[3].AsString);
            DataSet.Next;
          end;
      end
    else
      begin
        OpenSchema(siTables, EmptyParam, EmptyParam, DataSet);
        TypeField := DataSet.FieldByName('TABLE_TYPE'); { do not localize }
        NameField := DataSet.FieldByName('TABLE_NAME'); { do not localize }

        DataSet.First;
        while not DataSet.EOF do
          begin
            if not (itProcedure in which) then
              TableType := TypeField.AsString;
            if ((TableType = 'VIEW') and (itView in which)) or
              ((TableType = 'TABLE') and (itTable in which)) or
              ((TableType = 'SYSTEM TABLE') and (itSystem in which)) or
              (((TableType = 'LINK') or (TableType = 'PASS-THROUGH')) and (itLink in which)) then
              List.Add(NameField.AsString);
            DataSet.Next;
          end;
      end;
  finally
    DataSet.Free;
    List.EndUpdate;
  end;
end;

function TAJB_ADOConnection.GetModeString: string;
begin
  if ConnectionString <> '' then
    begin
      ConvertStringToList(ConnectionString, FConnectionList);
      Result := FConnectionList.Values['Data Source'];
    end
  else
    Result := '';
end;

function TAJB_ADOConnection.GetPassword: string;
begin
  if ConnectionString <> '' then
    begin
      ConvertStringToList(ConnectionString, FConnectionList);
      Result := FConnectionList.Values[cPassword];
    end
  else
    Result := '';
end;

procedure TAJB_ADOConnection.GetProcedureList(sl: TStrings);
begin
  getitemlist([itProcedure], sl);
end;

function TAJB_ADOConnection.GetProvider: TProvstring;
begin
  if ConnectionString <> '' then
    begin
      ConvertStringToList(ConnectionString, FConnectionList);
      Result := FConnectionList.Values['Provider'];
    end
  else
    Result := '';
end;

function TAJB_ADOConnection.GetQueryCommandText(const QueryName: string): string;
var
  i: Integer;
  v: View;
  p: Procedure_;
  cmd: OleVariant;
  done: Boolean;
  s: string;
  FCatalog: _Catalog;
begin
  done := false;
  Result := '';
  CheckActive;
  FCatalog := CoCatalog.Create;
  FCatalog._Set_ActiveConnection(ConnectionObject);
  try
    for i := 0 to FCatalog.Views.Count - 1 do
      begin
        s := FCatalog.Views.Item[i].Name;
        if (StrIComp(PChar(s), PChar(QueryName)) = 0) then
          done := True;
        v := FCatalog.Views.Item[i];
        Break;
      end;
    if done then
      begin
        cmd := v.Get_Command;
        Result := cmd.CommandText;
      end
    else
      begin
        for i := 0 to FCatalog.Procedures.Count - 1 do
          begin
            s := FCatalog.Procedures.Item[i].Name;
            if (StrIComp(PChar(s), PChar(QueryName)) = 0) then
              begin
                done := True;
                p := FCatalog.Procedures.Item[i];
                Break;
              end;
          end;
        if done then
          begin
            cmd := p.Get_Command;
            Result := cmd.commandtext;
          end;
      end;
  finally
    FCatalog := nil;
  end;
end;

procedure TAJB_ADOConnection.GetTableList(sl: TStrings);
begin
  getitemlist([itTable], sl);
end;

procedure TAJB_ADOConnection.GetTableNames(List: TStrings; SystemTables: Boolean = False);
//hides ancestor's GetTableNames
var
  tabset: TSchemaItemSet;
begin
  tabset := [itTable, itView, itLink];
  if SystemTables then
    Include(tabset, itSystem);
  getitemlist(tabset, List); //get all possible tables
end;

procedure TAJB_ADOConnection.GetUserList(sl: TStrings);
begin
  getitemlist([itUsers], sl);
end;

function TAJB_ADOConnection.GetVersion: string;
begin
  Result := versionStr;
end;

procedure TAJB_ADOConnection.GetViewList(sl: TStrings);
begin
  getitemlist([itView], sl);
end;

procedure TAJB_ADOConnection.MakeDatasetAs(Dest: TDataset; query: string);
//fills the fielddefs of dest with info from query
var
  adocols: TADOColumnList;
  oldactive: Boolean;
  i: Integer;
  p: PColInfo;
begin
  oldactive := Dest.Active;
  Dest.Active := False;
  Dest.FieldDefs.Clear;
  adocols := TADOColumnList.Create;
  try
    GetADOColumnlist(query, adocols);
    for i := 0 to adocols.Count - 1 do
      begin
        p := adocols[i];
        with Dest.FieldDefs.AddFieldDef do
          begin
            Name := p.FieldName;
            DataType := p.FieldType;
            Precision := p.FieldPrecision;
            Size := p.FieldSize;
          end;
      end;
  finally
    adocols.Free;
  end;
  Dest.Active := oldactive;
end;

procedure TAJB_ADOConnection.SetDatabaseName(const Value: TFilename);
begin
  if (ConnectionString <> '') and (FConnectionList.IndexOfName('Data Source') > -1) then
    FConnectionList.Values['Data Source'] := Value
  else
    FConnectionList.Add('Data Source=' + Value);
  ConnectionString := ConvertListtoString(FConnectionList);
end;

procedure TAJB_ADOConnection.SetEncrypt(Value: Boolean);
var
  s: string;
  x: Integer;
begin
  if Value then
    s := 'True'
  else
    s := 'False';
  x := FConnectionList.IndexOfName(cJETEncrypt);
  if (ConnectionString <> '') and (x > -1) then
    begin
      if not Value then
        FConnectionList.Delete(x)
      else
        FConnectionList.Values[cJETEncrypt] := s;
    end
  else if Value then
    FConnectionList.Add(cJETEncrypt + '=' + s);
  ConnectionString := ConvertListToString(FConnectionList);
end;

procedure TAJB_ADOConnection.SetISAM(Value: TISAMstring);
var
  i: Integer;
begin
  if Value = noISAM then
    begin
      i := FConnectionList.IndexOfName('Extended Properties');
      if i > -1 then FConnectionList.Delete(i);
    end
  else
    begin
      if (ConnectionString <> '') and (FConnectionList.IndexOfName('Extended Properties') > -1) then
        FConnectionList.Values['Extended Properties'] := Value
      else
        FConnectionList.Add('Extended Properties=' + Value);
      Provider := Jet4; //needed if ISAM is to be used!!
    end;
  ConnectionString := ConvertListtoString(FConnectionList);
end;

procedure TAJB_ADOConnection.SetModeString(const Value: string);
begin
  if (ConnectionString <> '') and (FConnectionList.IndexOfName('Mode') > -1) then
    FConnectionList.Values['Mode'] := Value
  else
    FConnectionList.Add('Mode=' + Value);
  ConnectionString := ConvertListtoString(FConnectionList);
end;

procedure TAJB_ADOConnection.SetPassword(const Value: string);
begin
  if (ConnectionString <> '') and (FConnectionList.IndexOfName(cPassword) > -1) then
    FConnectionList.Values[cPassword] := Value
  else
    FConnectionList.Add(cPassword + '=' + Value);
  ConnectionString := ConvertListToString(FConnectionList);
end;

procedure TAJB_ADOConnection.SetProvider(const Value: TProvstring);
begin
  if (ConnectionString <> '') and (FConnectionList.IndexOfName('Provider') > -1) then
    FConnectionList.Values['Provider'] := Value
  else
    FConnectionList.Add('Provider=' + Value);
  ConnectionString := ConvertListToString(FConnectionList);
end;

procedure TAJB_ADOConnection.SetVersion(const Value: string);
begin
  //do nothing
end;

function TAJB_ADOConnection.TableExists(const TableName: string): Boolean;
var
  i: Integer;
  s: string;
  FCatalog: _Catalog;
begin
  Result := false;
  CheckActive;
  FCatalog := CoCatalog.Create;
  FCatalog._Set_ActiveConnection(ConnectionObject);
  try
    FCatalog.Tables.Refresh;
    if not Result then
      for i := 0 to FCatalog.Tables.Count - 1 do
        begin
          s := FCatalog.Tables.Item[i].Name;
          if (StrIComp(PChar(s), PChar(TableName)) = 0) then
            begin
              Result := True;
              Break;
            end;
        end;

    FCatalog.Views.Refresh;
    if not Result then
      for i := 0 to FCatalog.Views.Count - 1 do
        begin
          s := FCatalog.Views.Item[i].Name;
          if (StrIComp(PChar(s), PChar(TableName)) = 0) then
            begin
              Result := True;
              Break;
            end;
        end;
  finally
    FCatalog := nil;
  end;
end;

destructor TADOColumnList.Destroy;
begin
  Clear;
  FList.Free;
  inherited;
end;

procedure TADOColumnList.Clear;
var
  i: Integer;
begin
  for i := 0 to FList.Count - 1 do
    Dispose(PColInfo(FList[i]));
  FList.Clear;
end;

function TADOColumnList.GetItems(Index: Integer): PColInfo;
begin
  if (Index >= 0) and (Index < Flist.Count) then
    Result := PColInfo(Flist[Index])
  else
    Result := nil;
end;

constructor TADOColumnList.Create;
begin
  Flist := TList.Create;
end;

function TADOColumnList.Add: PColInfo;
begin
  New(Result);
  FList.Add(Result);
end;

function TADOColumnList.GetCount: Integer;
begin
  Result := FList.Count;
end;

function TADOColumnList.GetFieldInfo(FieldName: string): PColInfo;
var
  i: Integer;
begin
  for i := 0 to FList.Count - 1 do
    if (CompareText(PColInfo(FList[i]).FieldName, FieldName) = 0) then
      Break;
  if (i < FList.Count) then
    Result := PColInfo(Flist[i])
  else
    Result := nil;
end;

constructor TODBC_Connection.Create(AOwner: TComponent);
begin
  inherited;
  LoginPrompt := False;
  inherited Provider := 'MSDASQL.1';
end;

destructor TODBC_Connection.Destroy;
begin
  inherited;
end;

function TODBC_Connection.GetProvider: WideString;
begin
  Result := 'MSDASQL.1';
end;

function TODBC_Connection.GetConnectionString: string;
begin
  Result := inherited ConnectionString;
end;

function TODBC_Connection.GetConnected: Boolean;
begin
  Result := inherited getConnected;
end;

procedure TODBC_Connection.SetConnected(Value: Boolean);
begin
  if Value then
    MakeConnectionString;
  inherited SetConnected(Value);
end;

procedure TODBC_Connection.MakeConnectionString;
const
  baseconnect = 'Provider=MSDASQL.1;Extended Properties="Driver={%s};';
var
  path, tempconnect: string;
begin
  path := ExcludeTrailingBackslash(ExtractFilePath(FDatabaseFile));
  DefaultDatabase := '';
  if (Pos('*.xls', FODBCDriver) > 0) then
    tempconnect := Format(baseconnect + 'DBQ=%s;"', [FODBCDriver, FDatabasefile])
  else if (Pos('*.dbf', FODBCDriver) > 0) then
    begin
      tempconnect := Format(baseconnect + '"', [FODBCDriver]);
      DefaultDatabase := path;
    end;
  inherited ConnectionString := tempconnect;
end;

end.

